(* module Engine
   This file provides two implementations of the interpreter for
   terms.

   VA is a substitute-and-repeat interpreter, similar to what Chris
   demonstrated for combinatory logic, only translated over to the
   more complex language we're working with here. The environments
   play no essential role in this version of the interpreter, which
   you can prove by specifying in types.ml that you want to use the
   Env0 or "do-nothing" implementation of environments.  (In that
   case, though, toplevel declarations like `let x = ... end` and
   `symbol` declarations will have no effect. But terms without any
   locally unbound variables, like `let y = 1 + 2 in let x = y + 10
   in (x, y)`, will still get interpreted via the same mechanisms
   and giving the same results.

   VB is an environment-based interpreter, and extends the same
   handling that VA uses for global/toplevel declarations also to
   the `y` and `x` in `let y = ... in let x = ... in (x,y)`. For
   this, you need to specify Env1 or Env2 in types.ml.

   To switch between these options, see the lines at the bottom
   of this file.
*)
  
open Lolevel.Error
open Interface
open Types
open Hilevel
open Hilevel.Printing
open Primitives


let die_with_stuck_term subterm =
  die (fun () -> begin
    newline();
    print_info (fst subterm);
    print "Stuck (sub)term: ";
    print_atom true subterm
  end)

let die_with_unbound_var fi var =
  let msg = Printf.sprintf "Unbound variable `%s`" in
  error_with fi (msg var)


module VA = struct
  (* ---------------------------------------------------------------
     Interpreter that reduces by substituting, using:
         val reduce_head_once : term -> env -> reduceOutcome
         val reduce : term -> env -> result
     `env` may contain toplevel bindings made by
     `let x = ... end;` or `symbol x;`

     reduce_head_once is similiar to reduce_if_redex in the
     combinatory logic interpreter, except that the latter only
     performed a reduction if its argument was exactly a redex.
     It had to rely on its caller to detect cases where the term
     was instead a longer sequence of applications that had a
     redex at its head.  In the present code, on the other hand,
     we have reduce_head_once take care of this itself, by
     calling itself recursively where appropriate. Still, it will
     only perform at most one reduction per invocation.
  *)
  

  (* Search term and find a variable that isn't used in it,
     either free or bound.  We do this by appending primes to base
     until we find an unused or "fresh" variable. *)
  let fresh_var (base : identifier) (term : term) : identifier =
    let rec aux term vars =
    match snd term with
    | TermVar(var_ident) ->
        var_ident :: vars
    | TermApp(head, arg) ->
        let vars' = aux head vars in
        aux arg vars'
    | TermLambda(bound_ident, body) ->
        aux body (bound_ident :: vars)
    | TermLet(bound_ident, arg, body) ->
        let vars' = aux arg vars in
        aux body (bound_ident :: vars')
    | TermLetRec _ -> failwith "letrec not implemented"
    | TermIf(test, yes, no) ->
        let vars' = aux test vars in
        let vars'' = aux yes vars' in
        aux no vars''
    | TermPrimfun _ | TermConstructor _ | TermDestructor _ | TermLiteral _ -> vars in
    let used_vars = aux term [] in
    let rec check ident =
      if List.mem ident used_vars
      then check (ident ^ "'")
      else ident in
    check (base ^ "'")
  
  let rec free_in (ident : identifier) (term : term) : bool =
    match term with
    | _,TermVar(var_ident) ->
        var_ident = ident
    | _,TermApp(head, arg) ->
        free_in ident head || free_in ident arg
    | _,TermLambda(bound_ident, body) ->
        bound_ident <> ident && free_in ident body
    | _,TermLet(bound_ident, arg, body) ->
        free_in ident arg || (bound_ident <> ident && free_in ident body)
    | _,TermLetRec _ -> failwith "letrec not implemented"
    | _,TermIf(test, yes, no) ->
        free_in ident test || free_in ident yes || free_in ident no
    | _,TermPrimfun _ | _,TermConstructor _ | _,TermDestructor _ | _,TermLiteral _ -> false
  
  let rec substitute (ident:identifier) (replacement : term) (original : term) =
    match original with
    | _,TermVar(var_ident) when var_ident = ident -> replacement
    | _,TermVar _ as orig -> orig
    | fi,TermApp(head, arg) ->
        let head' = substitute ident replacement head in
        let arg' = substitute ident replacement arg in
        fi,TermApp(head', arg')
    | _,TermLambda(bound_ident, body) as orig
      when bound_ident = ident || not (free_in ident body) ->
        (* vacuous substitution *)
        orig
    | fi,TermLambda(bound_ident, body)
      when not (free_in bound_ident replacement) ->
        (* can substitute without renaming bound_ident *)
        let body' = substitute ident replacement body in
        fi,TermLambda(bound_ident, body')
    | fi,TermLambda(bound_ident, body) ->
	(* find a fresh variable unused in either body or
           replacement (which we hack by specifying their App) *)
        let bound_ident' = fresh_var bound_ident (void_fi,TermApp(body,replacement)) in
        let body' = substitute bound_ident (void_fi,TermVar bound_ident') body in
        let body'' = substitute ident replacement body' in
        fi,TermLambda(bound_ident', body'')
    | fi,TermLet(bound_ident, arg, body)
      when bound_ident = ident || not (free_in ident body) ->
        let arg' = substitute ident replacement arg in
        fi,TermLet(bound_ident, arg', body)
    | fi,TermLet(bound_ident, arg, body)
      when not (free_in bound_ident replacement) ->
        (* can substitute without renaming bound_ident *)
        let body' = substitute ident replacement body in
        let arg' = substitute ident replacement arg in
        fi,TermLet(bound_ident, arg', body')
    | fi,TermLet(bound_ident, arg, body) ->
	(* find a fresh variable unused in either body or
           replacement (which we hack by specifying their App) *)
        let bound_ident' = fresh_var bound_ident (void_fi,TermApp(body,replacement)) in
        let body' = substitute bound_ident (void_fi,TermVar bound_ident') body in
        let body'' = substitute ident replacement body' in
        let arg' = substitute ident replacement arg in
        fi,TermLet(bound_ident', arg', body'')
    | _,TermLetRec _ -> failwith "letrec not implemented"
    | fi,TermIf(test, yes, no) ->
        let test' = substitute ident replacement test in
        let yes' = substitute ident replacement yes in
        let no' = substitute ident replacement no in
        fi,TermIf(test', yes', no')
    | _,TermPrimfun _ | _,TermConstructor _ | _,TermDestructor _ | _,TermLiteral _ as orig -> orig
  
  type reduceOutcome = AlreadyResult of result | StuckAt of term | ReducedTo of term
  
  let rec reduce_head_once (term : term) (env : env) : reduceOutcome =
    match term with
    | _,TermLambda _
    | _,TermPrimfun _
    | _,TermConstructor _
    | _,TermDestructor _
    | _,TermLiteral _ -> AlreadyResult (result_of_term term)
  
    | fi,TermVar var ->
        (match lookup var env with
        | Some (LocalBind _) -> assert false (* no LocalBinds in VA *)
        (* We have to return these as terms rather than results, else
	   the reduction rules for e.g. TermIf(Var("global"),...)
           will fail. *)
        | Some (TopBind res) -> ReducedTo (term_of_result res)
        | Some (SymbolBind) -> AlreadyResult (fi,Symbol var)
        | None -> die_with_unbound_var fi var)
  
    (* notice we never evaluate a yes/no branch until it is chosen *)
    | _,TermIf((_,TermLiteral(Bool true)),yes,no) ->
        ReducedTo yes
    | _,TermIf((_,TermLiteral(Bool false)),yes,no) ->
        ReducedTo no
    | fi,TermIf(test,yes,no) ->
        (match reduce_head_once test env with
        | AlreadyResult _ -> StuckAt term (* if test was not reducible to a bool, the if-term is not reducible at all *)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo test' -> ReducedTo(fi,TermIf(test',yes,no)))
  
    (* notice we never evaluate the body except after substituting
       and that happens only after arg is reduced to a result *)
    | fi,TermLet(bound_var,arg,body) ->
        (match reduce_head_once arg env with
        | AlreadyResult _ ->
            (* if arg was not reducible, we can substitute *)
            ReducedTo (substitute bound_var arg body)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo arg' -> ReducedTo (fi,TermLet(bound_var,arg',body)))
  
    | fi,TermLetRec(xts,t2) -> failwith "letrec not implemented"
  
    (* notice we only substitute after arg is reduced to a result *)
    | fi,TermApp(_,TermLambda(bound_var, body) as head, arg) ->
        (match reduce_head_once arg env with
        | AlreadyResult _ ->
            (* if arg was not reducible, we can substitute *)
            ReducedTo (substitute bound_var arg body)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo arg' -> ReducedTo (fi,TermApp(head, arg')))
  
    | fi,TermApp(head_fi,TermPrimfun(prim) as head, arg) ->
        (match reduce_head_once arg env with
        | AlreadyResult res ->
            (match apply_prim head_fi res prim with
            | Some res -> ReducedTo (term_of_result res)
            | None -> StuckAt term)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo arg' -> ReducedTo(fi,TermApp(head,arg')))
  
    | fi,TermApp(head_fi,TermConstructor(cons) as head, arg) ->
        (match reduce_head_once arg env with
        | AlreadyResult res ->
            (match apply_cons head_fi res cons with
            | Some res -> ReducedTo (term_of_result res)
            | None -> StuckAt term)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo arg' -> ReducedTo(fi,TermApp(head,arg')))
  
    | _,TermApp((fi,TermDestructor dest), (_,TermConstructor(Cons0(_,proj)))) ->
        (match proj dest with
        | Some res -> ReducedTo (term_of_result res)
        | None -> StuckAt term)
  
    (* first the head should be reduced, next the arg *)
    | fi,TermApp(head, arg) ->
        (match reduce_head_once head env with
        (* Literals and Symbols won't be reducible heads; for Symbols we'll
           reduce the args anyway. *)
        | AlreadyResult (_,ResLiteral _ as head') -> StuckAt (fi,TermApp(term_of_result head',arg))
        | AlreadyResult _ -> (* head was not reducible, was arg? *)
            (match reduce_head_once arg env with
            | ReducedTo arg' -> ReducedTo (fi,TermApp(head, arg'))
            (* reducible cases of App(result, result) were caught above; so here we're stuck *)
            | AlreadyResult _ -> StuckAt term
            | StuckAt _ as outcome -> outcome) (* propagate Stuck subterm *)
        | StuckAt _ as outcome -> outcome (* propagate Stuck subterm *)
        | ReducedTo head' -> ReducedTo (fi,TermApp(head', arg)))
  
  
  let rec reduce (term : term) (env : env) =
    match reduce_head_once term env with
    | AlreadyResult res -> res
    | StuckAt subterm -> die_with_stuck_term subterm
    | ReducedTo term' -> reduce term' env (* keep trying *)

end (* module VA *)

module VB = struct
  (* ---------------------------------------------------------------
     Interpreter that employs an "environment" or assignment
     function of results, using:
         val eval term -> env -> result, or may raise (Stuck term)
         val evaluate term -> env -> result
     Now `env` contains local as well as toplevel bindings.
  *)
  
  (*
     eval raises this exception when it fails to reduce/evaluate
     a term, because it has components for which no
     reduction/evaluation is defined, such as `x y` or `succ
     false`.  The substitute-and-repeat interpreter just signals
     this with a normal return value; but the environment-based
     interpreter uses an exception to abort prematurely.
  *)
  exception Stuck of term

  let rec eval (term : term) (env : env) : result =
    match term with
    | _,TermPrimfun _ | _,TermConstructor _ | _,TermDestructor _ | _,TermLiteral _ ->
        result_of_term term

    | fi,TermLambda(bound_var, body) -> fi,Closure(bound_var, body, env)

    | fi,TermVar var ->
        (match lookup var env with
        (* the first case is different from VA.reduce_head_once *)
        | Some (LocalBind res) -> res
        | Some (TopBind res) -> res
        | Some (SymbolBind) -> fi,Symbol var
        | None -> die_with_unbound_var fi var)

    | fi,TermIf(test, yes, no) ->
        (match eval test env with
        | _,ResLiteral(Bool true) -> eval yes env
        | _,ResLiteral(Bool false) -> eval no env
        (*  else test doesn't evaluate to a boolean *)
        | res -> raise (Stuck(fi,TermIf(term_of_result res,yes,no))))

    | _,TermLet(bound_var, arg, body) ->
	(* evaluate body under a new env where bound_var has been
	   bound to the result of evaluating arg under the
           current env *)
        let arg' = eval arg env in
        let env' = shift bound_var (LocalBind arg') env in
        eval body env'

    | fi,TermLetRec(xts,t2) -> failwith "letrec not implemented"

    | fi,TermApp(head, arg) ->
        (match eval head env with
        | _,Closure(bound_var, body, saved_env) ->
            (* argument gets evaluated in current env *)
            let arg' = eval arg env in
	    (* evaluate body under saved_env to govern its free
	       variables, except that we add a binding of
               bound_var to arg' *)
            let saved_env' = shift bound_var (LocalBind arg') saved_env in
            eval body saved_env'
        | head_fi,ResPrimfun(prim) as head' ->
            let arg' = eval arg env in
            (match apply_prim head_fi arg' prim with
            | Some res -> res
            | None -> raise (Stuck(fi,TermApp(term_of_result head',term_of_result arg'))))
        | head_fi,ResConstructor(cons) as head' ->
            let arg' = eval arg env in
            (match apply_cons head_fi arg' cons with
            | Some res -> res
            | None -> raise (Stuck(fi,TermApp(term_of_result head',term_of_result arg'))))
        | _,ResDestructor(dest) as head' ->
            (match eval arg env with
            | _,ResConstructor(Cons0(_,proj)) as arg' ->
              (match proj dest with
              | Some res -> res
              | None -> raise (Stuck(fi,TermApp(term_of_result head',term_of_result arg'))))
            | arg' -> raise (Stuck(fi,TermApp(term_of_result head',term_of_result arg'))))
        (* Literals and Symbols won't be reducible heads; but for Symbols we'll eval
           the arg anyway. This matches the behavior of reduce_head_once. *)
        | _,Symbol _ as head' ->
            let arg' = (try term_of_result(eval arg env) with Stuck term -> term) in
            raise (Stuck(fi,TermApp(term_of_result head',arg')))
        | head' -> raise (Stuck(fi,TermApp(term_of_result head',arg))))

  let evaluate (term : term) (env : env) : result =
    try eval term env
    with Stuck term -> die_with_stuck_term term

end (* module VB *)


(* Put comment (* *)s around exactly one of the following two pairs of lines. *)

let version = "A (reduce by substituting; " ^ version ^ ")"
let interpret = VA.reduce

(*
let version = "B (use environment for local bindings; " ^ version ^ ")"
let interpret = VB.evaluate
*)
